@@ -1,6 +1,6 @@
 #==============================================================================#
 #  Author:       Dominik Müller                                                #
-#  Copyright:    2021 IT-Infrastructure for Translational Medical Research,    #
+#  Copyright:    2022 IT-Infrastructure for Translational Medical Research,    #
 #                University of Augsburg                                        #
 #                                                                              #
 #  This program is free software: you can redistribute it and/or modify        #
@@ -28,14 +28,9 @@ import tempfile
 import pickle
 import os
 # Internal libraries
-from aucmedi.data_processing.io_data import image_loader
+from aucmedi.data_processing.io_loader import image_loader
 from aucmedi.data_processing.subfunctions import Standardize, Resize
 
-#-----------------------------------------------------#
-#                   Static Variables                  #
-#-----------------------------------------------------#
-ACCEPTABLE_IMAGE_FORMATS = ["jpeg", "jpg", "tif", "tiff", "png", "bmp", "gif"]
-
 #-----------------------------------------------------#
 #                 Keras Data Generator                #
 #-----------------------------------------------------#
@@ -49,8 +44,8 @@ ACCEPTABLE_IMAGE_FORMATS = ["jpeg", "jpg", "tif", "tiff", "png", "bmp", "gif"]
 
     The resulting batches are created based the following pipeline:
     - Image Loading
-    - Optional application of Data Augmentation
     - Optional application of Subfunctions
+    - Optional application of Data Augmentation
     - Standardize image
     - Stacking processed images to a batch
 
@@ -73,34 +68,41 @@ class DataGenerator(Iterator):
             labels (NumPy Array):           Classification list with One-Hot Encoding.
             image_format (String):          Image format to add at the end of the sample index for image loading.
             batch_size (Integer):           Number of samples inside a single batch.
-            resize (Tuple of Integers):     Resizing shape consisting of a X and Y size.
+            resize (Tuple of Integers):     Resizing shape consisting of a X and Y size. (optional Z size for Volumes)
+            subfunctions (List of Subfunctions):
+                                            List of Subfunctions class instances which will be SEQUENTIALLY executed on the data set.
             img_aug (ImageAugmentation):    Image Augmentation class instance which performs diverse data augmentation techniques.
             shuffle (Boolean):              Boolean, whether dataset should be shuffled.
             grayscale (Boolean):            Boolean, whether images are grayscale or RGB.
-            subfunctions (List of Subfunctions):
-                                            List of Subfunctions class instances which will be SEQUENTIALLY executed on the data set.
             standardize_mode (String):      Standardization modus in which image intensity values are scaled.
-            prepare_images (Boolean):       Boolean, whether all images should be prepared and backup to disk before training.
             sample_weights (List of Floats):List of weights for samples.
-            seed (Integer):                 Seed to ensure reproducibility for random function.
             workers (Integer):              Number of workers. If n_workers > 1 = use multi-threading for image preprocessing.
+            prepare_images (Boolean):       Boolean, whether all images should be prepared and backup to disk before training.
+                                            Recommended for large images or volumes to reduce CPU computing time.
+            loader (Function):              Function for loading samples/images from disk.
+            seed (Integer):                 Seed to ensure reproducibility for random function.
+            kwargs (Dictionary):            Additional parameters for the sample loader.
     """
     def __init__(self, samples, path_imagedir, labels=None, image_format=None,
-                 batch_size=32, resize=(224, 224), img_aug=None, shuffle=False,
-                 grayscale=False, subfunctions=[], standardize_mode="tf",
-                 prepare_images=False, sample_weights=None, seed=None,
-                 workers=1):
+                 batch_size=32, resize=(224, 224), subfunctions=[],
+                 img_aug=None, shuffle=False, grayscale=False,
+                 standardize_mode="z-score", sample_weights=None, workers=1,
+                 prepare_images=False,  loader=image_loader, seed=None,
+                 **kwargs):
         # Cache class variables
         self.labels = labels
         self.sample_weights = sample_weights
         self.prepare_images = prepare_images
         self.workers = workers
+        self.sample_loader = loader
+        self.kwargs = kwargs
         self.samples = samples
         self.path_imagedir = path_imagedir
         self.image_format = image_format
         self.grayscale = grayscale
         self.subfunctions = subfunctions
         self.img_aug = img_aug
+
         # Initialize Standardization Subfunction
         if standardize_mode is not None:
             self.sf_standardize = Standardize(mode=standardize_mode)
@@ -108,6 +110,9 @@ class DataGenerator(Iterator):
         # Initialize Resizing Subfunction
         if resize is not None : self.sf_resize = Resize(shape=resize)
         else : self.sf_resize = None
+        # Sanity check for full sample list
+        if samples is not None and len(samples) == 0:
+            raise ValueError("Provided sample list is empty!", len(samples))
         # Sanity check for label correctness
         if labels is not None and len(samples) != len(labels):
             raise ValueError("Samples and labels do not have same size!",
@@ -128,12 +133,20 @@ class DataGenerator(Iterator):
         if self.prepare_images:
             tmp_dir = tempfile.mkdtemp(prefix="aucmedi.tmp.", suffix=".data")
             self.prepare_dir = tmp_dir
-            for i in range(0, len(samples)):
-                preproc_img = self.preprocess_image(index=i,
-                                                    prepared_batch=False)
-                path_img = os.path.join(tmp_dir, "img_" + str(i))
-                with open(path_img + ".pickle", "wb") as pickle_writer:
-                    pickle.dump(preproc_img, pickle_writer)
+
+            # Preprocess image for each index - Sequential
+            if self.workers == 0 or self.workers == 1:
+                for i in range(0, len(samples)):
+                    self.preprocess_image(index=i, prepared_image=False,
+                                          run_aug=False, run_standardize=False,
+                                          dump_pickle=True)
+            # Preprocess image for each index - Multi-threading
+            else:
+                with ThreadPool(self.workers) as pool:
+                    index_array = list(range(0, len(samples)))
+                    mp_params = zip(index_array, repeat(False), repeat(False),
+                                    repeat(False), repeat(True))
+                    pool.starmap(self.preprocess_image, mp_params)
             print("A directory for image preparation was created:", tmp_dir)
 
         # Pass initialization parameters to parent Iterator class
@@ -154,7 +167,7 @@ class DataGenerator(Iterator):
         if self.workers == 0 or self.workers == 1:
             for i in index_array:
                 batch_img = self.preprocess_image(index=i,
-                                                  prepared_batch=self.prepare_images)
+                                                  prepared_image=self.prepare_images)
                 batch_stack[0].append(batch_img)
         # Process image for each index - Multi-threading
         else:
@@ -187,31 +200,47 @@ class DataGenerator(Iterator):
     """Preprocessing function for applying subfunctions, augmentation, resizing and standardization
        on an image given its index.
 
-       Activating the prepared_batch option also allows loading a beforehand preprocessed image from disk.
+       Activating the prepared_image option also allows loading a beforehand preprocessed image from disk.
+       Deactivating the run_aug & run_standardize option to output image without augmentation and standardization.
+       Activating dump_pickle will store the preprocessed image as pickle on disk instead of returning.
     """
-    def preprocess_image(self, index, prepared_batch=False):
+    def preprocess_image(self, index, prepared_image=False, run_aug=True,
+                         run_standardize=True, dump_pickle=False):
         # Load prepared image from disk
-        if prepared_batch:
+        if prepared_image:
+            # Load from disk
             path_img = os.path.join(self.prepare_dir, "img_" + str(index))
             with open(path_img + ".pickle", "rb") as pickle_loader:
                 img = pickle.load(pickle_loader)
+            # Apply image augmentation on image if activated
+            if self.img_aug is not None and run_aug:
+                img = self.img_aug.apply(img)
+            # Apply standardization on image if activated
+            if self.sf_standardize is not None and run_standardize:
+                img = self.sf_standardize.transform(img)
         # Preprocess image during runtime
         else:
-            # Load image
-            img = image_loader(self.samples[index], self.path_imagedir,
-                               image_format=self.image_format,
-                               grayscale=self.grayscale)
-            # Apply image augmentation on image if activated
-            if self.img_aug is not None:
-                img = self.img_aug.apply(img)
+            # Load image from disk
+            img = self.sample_loader(self.samples[index], self.path_imagedir,
+                                     image_format=self.image_format,
+                                     grayscale=self.grayscale,
+                                     **self.kwargs)
             # Apply subfunctions on image
             for sf in self.subfunctions:
                 img = sf.transform(img)
             # Apply resizing on image if activated
             if self.sf_resize is not None:
                 img = self.sf_resize.transform(img)
+            # Apply image augmentation on image if activated
+            if self.img_aug is not None and run_aug:
+                img = self.img_aug.apply(img)
             # Apply standardization on image if activated
-            if self.sf_standardize is not None:
+            if self.sf_standardize is not None and run_standardize:
                 img = self.sf_standardize.transform(img)
+        # Dump preprocessed image to disk (for later usage via prepared_image)
+        if dump_pickle:
+            path_img = os.path.join(self.prepare_dir, "img_" + str(index))
+            with open(path_img + ".pickle", "wb") as pickle_writer:
+                pickle.dump(img, pickle_writer)
         # Return preprocessed image
-        return img
+        else : return img
